-- 
-- hsbirthdays-@VERSION@
-- (c) 2005 Andres Loeh <hsbirthdays at andres-loeh dot de>
-- distributed under the GPL-2
--

import System.Time
import System.Environment
import Control.Monad.State
import Data.List
import Debug.Trace
import System.Console.GetOpt

progname = "hsbirthdays"

data Birthday  =  B  Bool -- has year?
                     CalendarTime -- when?
                     String -- text

type Errors = [String]

data Opts = O  {  files  ::  [String],
                  days   ::  Int,
                  age    ::  Bool,
                  years  ::  Bool  }

defaults = 
   do  home <- getEnv "HOME"
       return $ O { files = [home ++ "/." ++ progname], days = 28, age = True, years = True }

options = [  Option "f" ["file"] (ReqArg (\s o -> o { files = s : files o }) "name")  "specify birthdays file(s)",
             Option "d" ["days"] (ReqArg (\d o -> o { days = (read d) }) "number")    "how many days",
             Option []  ["noage"] (NoArg (\o -> o { age = False }))                   "don't show age",
             Option []  ["noyears"] (NoArg (\o -> o { years = False }))               "don't show year numbers" ]
                                

main =
   do  dopts <- defaults
       args <- getArgs
       let (opts,errs,nopts) = getOpt Permute options args
       if not (null (errs ++ nopts))
          then  putStrLn $ usageInfo ("hsbirthdays, version @VERSION@\nusage: " ++ progname ++ " [options]") options
          else  proceed (foldl (flip ($)) dopts opts)

proceed opts =
   do  (birthdays,errors) <- parseFiles $ files opts
       mapM_ putStrLn errors
       time <- getClockTime
       ctime <- toCalendarTime time
       let year = ctYear ctime
       -- real times
       let rt = concatMap (realTime year) birthdays
       -- select timespan of interest
       let sel = filter (\(x,_) ->  let diff   =  normalizeTimeDiff $ diffClockTimes (toClockTime x) time
                                        rdiff  =  normalizeTimeDiff $ noTimeDiff { tdDay = days opts }
                                        ndiff  =  normalizeTimeDiff $ noTimeDiff { tdDay = -1 } -- we want todays
                                    in  ndiff < diff && diff <= rdiff) rt
       mapM_ putStrLn $ map (uncurry (printBirthday opts)) (sortBy (\(x,_) (y,_) -> compare x y) sel)

parseFiles f =
   do  xs <- readFiles f
       let  entries = lines xs  -- one line per birthday
            nrentries = zip [1..] entries -- line numbers
            stripped = filter (\(_,xs) -> case xs of 
                                            [] -> False
                                            ('#':_) -> False
                                            _ -> True) nrentries
                                            -- filter blank lines and comments
            split = map (\(x,y) -> (x,words y)) stripped
            (birthdayss,errors) = runState (mapM parseBirthday split) []
       return (concat birthdayss,errors)

-- the last entry is the default; if user-specified files are present, ignore it
readFiles [f]  =  readFile f
readFiles fs   =  mapM readFile (init fs) >>= return . concat

addError :: String -> State Errors ()
addError n =  modify (n:)

realTime :: Int -> Birthday -> [(CalendarTime,Birthday)]
realTime y b@(B _ t _) = [(t { ctYear = y },b), (t { ctYear = y+1 },b)]

printBirthday :: Opts -> CalendarTime -> Birthday -> String
printBirthday opts t (B b o txt) =
   let  d = ctDay t
        m = fromEnum (ctMonth t) + 1
        y = ctYear t
        date = formatTwo d ++ "." ++ formatTwo m ++ "."
	       ++ if years opts then show y else ""
   in   date ++ " " ++ txt ++
        if b && age opts then " (" ++ show (y - ctYear o) ++ ")" else ""

formatTwo :: Int -> String
formatTwo n = (if n < 10 then "0" else "") ++ show n
  

parseBirthday :: (Int,[String]) -> State Errors [Birthday]
parseBirthday (n,d:txt@(_:_)) 
                        = do  x <- parseDate n d
                              case x of
                                Nothing      ->  return []
                                Just (y,pd)  ->  return $ [B y pd (unwords txt)]
parseBirthday (n,_)     = do  addError $ show n ++ ": malformed line"
                              return []

parseDate :: Int -> String -> State Errors (Maybe (Bool,CalendarTime))
parseDate n d = 
   do  is <- parseDotSequence n d
       case is of
         [d,m,y]  ->  return (Just (True,  emptyTime {  ctDay    =  d,
                                                        ctMonth  =  toEnum (m-1),
                                                        ctYear   =  y  }))
         [d,m]    ->  return (Just (False, emptyTime {  ctDay    =  d,
                                                        ctMonth  =  toEnum (m-1)  }))
         _        ->  do  addError $ show n ++ ": malformed date"
                          return Nothing

emptyTime = toUTCTime (TOD 0 0)

parseDotSequence :: Int -> String -> State Errors [Int]
parseDotSequence n s = 
   case findIndex (=='.') s of 
     Nothing -> case s of
                  [] -> return []
                  _  -> case reads s of
                          [(i,"")] -> return [i]
                          _ -> do  addError $ show n ++ ": malformed date"
                                   return []
     Just n  -> case splitAt n s of
                  (pre,_:post) -> case reads pre of
                                    [(i,"")] -> do  rs <- parseDotSequence n post
                                                    return (i : rs)
                                    _ -> do  addError $ show n ++ ": malformed date"
                                             return []
